/******************************************************************************
* Copyright (c) 2015, Hobu Inc. (info@hobu.co)
*
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following
* conditions are met:
*
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in
*       the documentation and/or other materials provided
*       with the distribution.
*     * Neither the name of Hobu, Inc. or Flaxen Geo Consulting nor the
*       names of its contributors may be used to endorse or promote
*       products derived from this software without specific prior
*       written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
* "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
* LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
* FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
* COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
* INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
* BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
* OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
* AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
* OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
* OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
* OF SUCH DAMAGE.
****************************************************************************/

#include "MergeKernel.hpp"

#include <filters/MergeFilter.hpp>
#include <pdal/StageFactory.hpp>

namespace pdal
{

static StaticPluginInfo const s_info
{
    "kernels.merge",
    "Merge Kernel",
    "https://pdal.org/apps/merge.html"
};

CREATE_STATIC_KERNEL(MergeKernel, s_info)

std::string MergeKernel::getName() const
{
    return s_info.name;
}


void MergeKernel::addSwitches(ProgramArgs& args)
{
    args.add("files,f", "input/output files", m_files).setPositional();
}


void MergeKernel::validateSwitches(ProgramArgs& args)
{
    if (m_files.size() < 2)
        throw pdal_error("Must specify an input and output file.");
    m_outputFile = m_files.back();
    m_files.resize(m_files.size() - 1);
}


int MergeKernel::execute()
{
    MergeFilter filter;

    for (size_t i = 0; i < m_files.size(); ++i)
    {
        Stage& reader = makeReader(m_files[i], m_driverOverride);
        filter.setInput(reader);
    }

    Stage& writer = makeWriter(m_outputFile, filter, "");

    ColumnPointTable table;
    writer.prepare(table);
    writer.execute(table);
    return 0;
}

} // namespace pdal
